/*
 * yasapi_time.c
 * Copyright (C) 2015-2016 Peter Belkner <pbelkner@snafu.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <yasapi.h>

int ConnectionGetPosition(Connection *pConnect, UINT64 *pu64Position)
{
  IAudioClock *pClock=pConnect->pClock;
  HRESULT hr;

  if (pConnect->bInvalid||!pClock)
    goto invalid;

  // get position /////////////////////////////////////////////////////////////
  hr=pClock->lpVtbl->GetPosition(pClock,
    pu64Position,         // [out] UINT64 *pu64Position,
    NULL                  // [out] UINT64 *pu64QPCPosition
  );

  if (FAILED(hr)) {
    if (hr==AUDCLNT_E_DEVICE_INVALIDATED) {
      pConnect->bInvalid=1;
      goto invalid;
    }

    DERROR(E_POINTER,hr);
    DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(hr);
    DMESSAGE("getting position");
    goto position;
  }

  return 0;
position:
invalid:
  return -1;
}

int ConnectionGetFrequency(Connection *pConnect, UINT64 *pu64Frequency)
{
  IAudioClock *pClock=pConnect->pClock;
  HRESULT hr;

  if (pConnect->bInvalid||!pClock)
    goto invalid;

  // get frequency ////////////////////////////////////////////////////////////
  hr=pClock->lpVtbl->GetFrequency(pClock,
    pu64Frequency         // [out] UINT64 *pu64Frequency
  );

  if (FAILED(hr)) {
    if (hr==AUDCLNT_E_DEVICE_INVALIDATED) {
      pConnect->bInvalid=1;
      goto invalid;
    }

    DERROR(E_POINTER,hr);
    DERROR(AUDCLNT_E_DEVICE_INVALIDATED,hr);
    DERROR(AUDCLNT_E_SERVICE_NOT_RUNNING,hr);
    DUNKNOWN(hr);
    DMESSAGE("getting frequency");
    goto frequency;
  }
  else if (*pu64Frequency==0) {
    DMESSAGE("zero frequency");
    goto zero;
  }

  return 0;
zero:
frequency:
invalid:
  return -1;
}

void TimeSegmentReset(TimeSegment *pSegment, UINT64 u64Frequency)
{
  if (u64Frequency)
    pSegment->u64Position=0ll;
  else
    pSegment->xSeconds=0.0;
}

void TimeSegmentAdd(TimeSegment *pOffset, TimeSegment *pCurrent,
    UINT64 u64Frequency)
{
  if (u64Frequency)
    pOffset->u64Position+=pCurrent->u64Position;
  else
    pOffset->xSeconds+=pCurrent->xSeconds;
}

#if 0 // {
void TimeSegmentSub(TimeSegment *pOffset, TimeSegment *pCurrent,
    UINT64 u64Frequency)
{
  if (u64Frequency)
    pOffset->u64Position-=pCurrent->u64Position;
  else
    pOffset->xSeconds-=pCurrent->xSeconds;
}
#endif // }

int TimeSegmentGet(TimeSegment *pSegment, UINT64 u64Frequency,
    Connection *pConnect)
{
  UINT64 u64Position;

  if (pConnect->bInvalid||!pConnect->pClock)
    goto invalid;

  if (ConnectionGetPosition(pConnect,&u64Position)<0) {
    DMESSAGE("getting position");
    goto position;
  }

  if (u64Frequency)
    pSegment->u64Position=u64Position;
  else {
    if (ConnectionGetFrequency(pConnect,&u64Frequency)<0) {
      DMESSAGE("getting frequency");
      goto frequency;
    }

    pSegment->xSeconds=(double)u64Position/u64Frequency;
  }

  return 0;
frequency:
position:
invalid:
  return -1;
}

int TimeReset(Time *pTime, TimeTag eTimeTag, Connection *pConnect)
{
  enum { DEBUG=0 };

  DPRINTF(DEBUG,"  > %s <\n",__func__);

  if (pConnect->bInvalid||!pConnect->pClock)
    goto invalid;

  switch (eTimeTag) {
  case TIME_POSITION:
    if (pConnect->bInvalid)
      goto invalid;
    else if (ConnectionGetFrequency(pConnect,&pTime->u64Frequency)<0) {
      DMESSAGE("getting frequency");
      goto frequency;
    }

    break;
  case TIME_TIME:
    pTime->u64Frequency=0ull;
    break;
  default:
    DMESSAGEV("tag mismatch: %d\n",eTimeTag);
    goto tag;
  }

  TimeSegmentReset(&pTime->gapless,pTime->u64Frequency);
  TimeSegmentReset(&pTime->pause,pTime->u64Frequency);
  TimeSegmentReset(&pTime->current,pTime->u64Frequency);

  return 0;
tag:
frequency:
invalid:
  return -1;
}

void TimeFlush(Time *pTime)
{
  DPRINTF(0,"  > %s <\n",__func__);

  TimeSegmentReset(&pTime->pause,pTime->u64Frequency);
  TimeSegmentReset(&pTime->current,pTime->u64Frequency);
}

int TimePause(Time *pTime, Connection *pConnect)
{
  DPRINTF(0,"  > %s <\n",__func__);

  if (pConnect->bInvalid||!pConnect->pClock)
    goto invalid;

  if (TimeSegmentGet(&pTime->current,pTime->u64Frequency,pConnect)<0) {
    DMESSAGE("getting time");
    goto time;
  }

  TimeSegmentReset(&pTime->gapless,pTime->u64Frequency);
  pTime->pause=pTime->current;

  return 0;
time:
invalid:
  return -1;
}

int TimeGapless(Time *pTime, Connection *pConnect)
{
  DPRINTF(0,"  > %s <\n",__func__);

  if (pConnect->bInvalid||!pConnect->pClock)
    goto invalid;

  TimeSegmentReset(&pTime->pause,pTime->u64Frequency);

  if (TimeSegmentGet(&pTime->gapless,pTime->u64Frequency,pConnect)<0) {
    DMESSAGE("getting time");
    goto time;
  }

  pTime->current=pTime->gapless;

  return 0;
time:
invalid:
  return -1;
}

int TimeMigrate(Time *pTime, Connection *pConnect)
{
  DPRINTF(0,"  > %s <\n",__func__);

  if (pConnect->bInvalid||!pConnect->pClock)
    goto invalid;

  TimeSegmentAdd(&pTime->pause,&pTime->current,pTime->u64Frequency);
  TimeSegmentReset(&pTime->current,pTime->u64Frequency);

  return 0;
time:
invalid:
  return -1;
}

int TimeGetMS(Time *pTime, Connection *pConnect, double *pms)
{
  enum { DEBUG=3 };
  TimeSegment *pCurrent=&pTime->current;
  TimeSegment *pPause=&pTime->pause;
  TimeSegment *pGapless=&pTime->gapless;

  if (!pConnect||pConnect->bInvalid||!pConnect->pClock)
    TimeSegmentReset(pCurrent,pTime->u64Frequency);
  else if (TimeSegmentGet(pCurrent,pTime->u64Frequency,pConnect)<0) {
    DMESSAGE("getting current time\n");
    goto time;
  }

  if (pTime->u64Frequency) {
    if (!pTime->u64Frequency) {
      DMESSAGE("division by zero frequency\n");
      goto frequency;
    }

    *pms=1000.0*(pCurrent->u64Position+pPause->u64Position
        -pGapless->u64Position)/pTime->u64Frequency;
  }
  else
    *pms=1000.0*(pCurrent->xSeconds+pPause->xSeconds
        -pGapless->xSeconds);

  DPRINTF(DEBUG,"  %s: %.0f\n",__func__,*pms+0.5);

  return 0;
frequency:
time:
  return -1;
}
